@@ -55,10 +55,56 @@ namespace dlib
                 - returns 1
     !*/
 
+    template <typename T>
+    void set_learning_rate_multiplier(
+        T& obj,
+        double learning_rate_multiplier
+    );
+    /*!
+        requires
+            - learning_rate_multiplier >= 0
+        ensures
+            - if (obj has a set_learning_rate_multiplier() member function) then
+                - calls obj.set_learning_rate_multiplier(learning_rate_multiplier)
+            - else
+                - does nothing
+    !*/
+
+// ----------------------------------------------------------------------------------------
+
+    template <typename T>
+    double get_bias_learning_rate_multiplier(
+        const T& obj
+    );
+    /*!
+        ensures
+            - if (obj has a get_bias_learning_rate_multiplier() member function) then
+                - returns obj.get_bias_learning_rate_multiplier()
+            - else
+                - returns 1
+    !*/
+
+    template <typename T>
+    void set_bias_learning_rate_multiplier(
+        T& obj,
+        double bias_learning_rate_multiplier
+    );
+    /*!
+        requires
+            - bias_learning_rate_multiplier >= 0
+        ensures
+            - if (obj has a set_bias_learning_rate_multiplier() member function) then
+                - calls obj.set_bias_learning_rate_multiplier(bias_learning_rate_multiplier)
+            - else
+                - does nothing
+    !*/
+
+// ----------------------------------------------------------------------------------------
+
     template <typename T>
     double get_weight_decay_multiplier(
         const T& obj
-    ); 
+    );
     /*!
         ensures
             - if (obj has a get_weight_decay_multiplier() member function) then
@@ -67,6 +113,64 @@ namespace dlib
                 - returns 1
     !*/
 
+    template <typename T>
+    void set_weight_decay_multiplier(
+        T& obj,
+        double weight_decay_multiplier
+    );
+    /*!
+        requires
+            - weight_decay_multiplier >= 0
+        ensures
+            - if (obj has a set_weight_decay_multiplier() member function) then
+                - calls obj.set_weight_decay_multiplier(weight_decay_multiplier)
+            - else
+                - does nothing
+    !*/
+
+// ----------------------------------------------------------------------------------------
+
+    template <typename T>
+    double get_bias_weight_decay_multiplier(
+        const T& obj
+    ); 
+    /*!
+        ensures
+            - if (obj has a get_bias_weight_decay_multiplier() member function) then
+                - returns obj.get_bias_weight_decay_multiplier()
+            - else
+                - returns 1
+    !*/
+
+    template <typename T>
+    void set_bias_weight_decay_multiplier(
+        T& obj,
+        double bias_weight_decay_multiplier
+    );
+    /*!
+        requires:
+            - bias_weight_decay_multiplier >= 0
+        ensures
+            - if (obj has a set_bias_weight_decay_multiplier() member function) then
+                - calls obj.set_bias_weight_decay_multiplier(bias_weight_decay_multiplier)
+            - else
+                - does nothing
+    !*/
+
+// ----------------------------------------------------------------------------------------
+
+    template <typename T>
+    void disable_bias(
+        T& obj
+    );
+    /*!
+        ensures
+            - if (obj has a disable_bias() member function) then
+                - calls obj.disable_bias()
+            - else
+                - does nothing
+    !*/
+
 // ----------------------------------------------------------------------------------------
 
     bool dnn_prefer_fastest_algorithms(
@@ -549,6 +653,13 @@ namespace dlib
                 - The solvers use the given learning rate.
         !*/
 
+        template <typename solver_type>
+        void update_parameters(std::vector<solver_type>& solvers, double learning_rate)
+        { update_parameters(make_sstack(solvers), learning_rate); }
+        /*!
+            Convenience method for calling update_parameters()
+        !*/
+
         void clean(
         );
         /*!
@@ -760,6 +871,29 @@ namespace dlib
 
     // -------------
 
+        const tensor& forward(const tensor& x
+        ); 
+        /*!
+            requires
+                - sample_expansion_factor() != 0
+                  (i.e. to_tensor() must have been called to set sample_expansion_factor()
+                  to something non-zero.)
+                - x.num_samples()%sample_expansion_factor() == 0
+                - x.num_samples() > 0
+            ensures
+                - Runs x through the network and returns the results as a tensor.  In particular,
+                  this function just performs:
+                    return subnet().forward(x);
+                  So if you want to get the outputs as an output_label_type then call one of the
+                  methods below instead, like operator().
+                - The return value from this function is also available in #subnet().get_output().
+                  i.e. this function returns #subnet().get_output().
+                - have_same_dimensions(#subnet().get_gradient_input(), #subnet().get_output()) == true
+                - All elements of #subnet().get_gradient_input() are set to 0. 
+                  i.e. calling this function clears out #subnet().get_gradient_input() and ensures
+                  it has the same dimensions as the most recent output.
+        !*/
+
         template <typename output_iterator>
         void operator() (
             const tensor& x, 
@@ -899,6 +1033,9 @@ namespace dlib
                 - for all valid k:
                     - the expected label of the kth sample in x is *(lbegin+k/sample_expansion_factor()).
                 - This function does not update the network parameters.
+                - For sub-layers that are immediate inputs into the loss layer, we also populate the
+                  sub-layer's get_gradient_input() tensor with the gradient of the loss with respect
+                  to the sub-layer's output.
         !*/
 
         template <typename forward_iterator, typename label_iterator>
@@ -919,6 +1056,9 @@ namespace dlib
                 - for all valid k:
                     - the expected label of *(ibegin+k) is *(lbegin+k).
                 - This function does not update the network parameters.
+                - For sub-layers that are immediate inputs into the loss layer, we also populate the
+                  sub-layer's get_gradient_input() tensor with the gradient of the loss with respect
+                  to the sub-layer's output.
         !*/
 
     // -------------
@@ -937,6 +1077,9 @@ namespace dlib
             ensures
                 - runs x through the network and returns the resulting loss. 
                 - This function does not update the network parameters.
+                - For sub-layers that are immediate inputs into the loss layer, we also populate the
+                  sub-layer's get_gradient_input() tensor with the gradient of the loss with respect
+                  to the sub-layer's output.
         !*/
 
         template <typename forward_iterator>
@@ -952,6 +1095,9 @@ namespace dlib
             ensures
                 - runs [ibegin,iend) through the network and returns the resulting loss. 
                 - This function does not update the network parameters.
+                - For sub-layers that are immediate inputs into the loss layer, we also populate the
+                  sub-layer's get_gradient_input() tensor with the gradient of the loss with respect
+                  to the sub-layer's output.
         !*/
 
     // -------------
@@ -1065,6 +1211,73 @@ namespace dlib
                 - The solvers use the given learning rate.
         !*/
 
+        template <typename solver_type>
+        void update_parameters(std::vector<solver_type>& solvers, double learning_rate
+        ) { update_parameters(make_sstack(solvers), learning_rate); }
+        /*!
+            Convenience method for calling update_parameters()
+        !*/
+
+        void back_propagate_error(
+            const tensor& x
+        );
+        /*!
+            requires
+                - forward(x) was called to forward propagate x though the network.
+                  Moreover, this was the most recent call to forward() and x has not been
+                  subsequently modified in any way.
+                - subnet().get_gradient_input() has been set equal to the gradient of this network's
+                  output with respect to the loss function (generally this will be done by calling
+                  compute_loss()).
+            ensures
+                - Back propagates the error gradient, subnet().get_gradient_input(), through this
+                  network and computes parameter and data gradients, via backpropagation.
+                  Specifically, this function populates get_final_data_gradient() and also,
+                  for each layer, the tensor returned by get_parameter_gradient().
+                - All elements of #subnet().get_gradient_input() are set to 0. 
+                - have_same_dimensions(#get_final_data_gradient(), x) == true.
+                - #get_final_data_gradient() contains the gradient of the network with
+                  respect to x.
+        !*/
+
+        void back_propagate_error(
+            const tensor& x, 
+            const tensor& gradient_input
+        );
+        /*!
+            requires
+                - forward(x) was called to forward propagate x though the network.
+                  Moreover, this was the most recent call to forward() and x has not been
+                  subsequently modified in any way.
+                - have_same_dimensions(gradient_input, subnet().get_output()) == true
+            ensures
+                - This function is identical to the version of back_propagate_error()
+                  defined immediately above except that it back-propagates gradient_input
+                  through the network instead of subnet().get_gradient_input().  Therefore, this
+                  version of back_propagate_error() is equivalent to performing:
+                    subnet().get_gradient_input() = gradient_input;
+                    back_propagate_error(x);
+                  Except that calling back_propagate_error(x,gradient_input) avoids the
+                  copy and is therefore slightly more efficient.
+                - All elements of #subnet.get_gradient_input() are set to 0. 
+                - have_same_dimensions(#get_final_data_gradient(), x) == true.
+                - #get_final_data_gradient() contains the gradient of the network with
+                  respect to x.
+        !*/
+
+        const tensor& get_final_data_gradient(
+        ) const; 
+        /*!
+            ensures
+                - if back_propagate_error() has been called to back-propagate a gradient
+                  through this network then you can call get_final_data_gradient() to
+                  obtain the last data gradient computed.  That is, this function returns
+                  the gradient of the network with respect to its inputs.
+                - Note that there is only one "final data gradient" for an entire network,
+                  not one per layer, since there is only one input to the entire network.
+        !*/
+
+
     // -------------
 
         void clean (
@@ -1463,6 +1676,8 @@ namespace dlib
               add_tag_layer.
             - v is a function object with a signature equivalent to: 
                 v(size_t idx, tensor& t)
+              or:
+                v(tensor& t)
         ensures
             - Loops over all the computational layers (i.e. layers with parameters, as
               opposed to loss, tag, or input layers) in net and passes their parameters to
@@ -1496,6 +1711,8 @@ namespace dlib
               add_tag_layer.
             - v is a function object with a signature equivalent to: 
                 v(size_t idx, tensor& t)
+              or:
+                v(tensor& t)
         ensures
             - Loops over all the computational layers (i.e. layers with parameters, as
               opposed to loss, tag, or input layers) in net and passes their parameter
@@ -1530,7 +1747,9 @@ namespace dlib
               add_tag_layer.
             - v is a function object with a signature equivalent to: 
                 v(size_t idx, any_net_type& t)
-              That is, it must take a size_t and then any of the network types such as
+              or:
+                v(any_net_type& t)
+              That is, it takes an optional size_t and then any of the network types such as
               add_layer, add_loss_layer, etc.
         ensures
             - Loops over all the layers in net and calls v() on them.  To be specific, this
@@ -1554,7 +1773,9 @@ namespace dlib
               add_tag_layer.
             - v is a function object with a signature equivalent to: 
                 v(size_t idx, any_net_type& t)
-              That is, it must take a size_t and then any of the network types such as
+              or:
+                v(any_net_type& t)
+              That is, it takes an optional size_t and then any of the network types such as
               add_layer, add_loss_layer, etc.
         ensures
             - Loops over all the layers in net and calls v() on them.  The loop happens in
@@ -1565,6 +1786,64 @@ namespace dlib
                     v(i-1, layer<i-1>(net));
     !*/
 
+// ----------------------------------------------------------------------------------------
+
+    template <
+        typename net_type,
+        typename visitor
+        >
+    void visit_computational_layers(
+        net_type& net,
+        visitor v
+    );
+    /*!
+        requires
+            - net_type is an object of type add_layer, add_loss_layer, add_skip_layer, or
+              add_tag_layer.
+            - v is a function object with a signature equivalent to: 
+                v(size_t idx, any_computational_layer& t)
+              or:
+                v(any_computational_layer& t)
+              That is, it takes an optional size_t and then any of the computational layers.  E.g.
+              one of the layer types defined in dlib/dnn/layers_abstract.h like fc_ or conv_.
+        ensures
+            - Loops over all the computational layers in net and calls v() on them.  To be specific, this
+              function essentially performs the following:
+
+                for (size_t i = 0; i < net_type::num_layers; ++i)
+                    if (layer<i>(net) is an add_layer type, i.e. it adds a computational layer)
+                        v(i, layer<i>(net).layer_details());
+    !*/
+
+    template <
+        size_t begin,
+        size_t end,
+        typename net_type,
+        typename visitor
+        >
+    void visit_computational_layers_range(
+        net_type& net,
+        visitor v
+    );
+    /*!
+        requires
+            - net_type is an object of type add_layer, add_loss_layer, add_skip_layer, or
+              add_tag_layer.
+            - v is a function object with a signature equivalent to: 
+                v(size_t idx, any_computational_layer& t)
+              or:
+                v(any_computational_layer& t)
+              That is, it takes an optional size_t and then any of the computational layers.  E.g.
+              one of the layer types defined in dlib/dnn/layers_abstract.h like fc_ or conv_.
+        ensures
+            - Loops over all the computational layers in the range [begin,end) in net and calls v()
+              on them.  To be specific, this function essentially performs the following:
+
+                for (size_t i = begin; i < end; ++i)
+                    if (layer<i>(net) is an add_layer type, i.e. it adds a computational layer)
+                        v(i, layer<i>(net).layer_details());
+    !*/
+
 // ----------------------------------------------------------------------------------------
 
     template <
@@ -1583,13 +1862,14 @@ namespace dlib
               add_tag_layer.
             - v is a function object with a signature equivalent to: 
                 v(size_t idx, any_net_type& t)
-              That is, it must take a size_t and then any of the network types such as
+              or:
+                v(any_net_type& t)
+              That is, it takes an optional size_t and then any of the network types such as
               add_layer, add_loss_layer, etc.
             - begin <= end <= net_type::num_layers
         ensures
             - Loops over the layers in the range [begin,end) in net and calls v() on them.
-              The loop happens in the reverse order of visit_layers().  To be specific,
-              this function essentially performs the following:
+              To be specific, this function essentially performs the following:
 
                 for (size_t i = begin; i < end; ++i)
                     v(i, layer<i>(net));
@@ -1611,7 +1891,9 @@ namespace dlib
               add_tag_layer.
             - v is a function object with a signature equivalent to: 
                 v(size_t idx, any_net_type& t)
-              That is, it must take a size_t and then any of the network types such as
+              or:
+                v(any_net_type& t)
+              That is, it takes an optional size_t and then any of the network types such as
               add_layer, add_loss_layer, etc.
             - begin <= end <= net_type::num_layers
         ensures
